########## R script: growthIndianaExample ##########

# For performing the example in the Menictas, Nolan, Simpson & Wand 
# article in which two-level group specific curves, with a contrast
# curve extension, are fitted to data from a longitudinal study on 
# adolescent somatic growth from the Indiana University School of 
# Medicine, Indianapolis, Indiana, U.S.A. using a streamlined
# Bayesian variational inference approach.

# Last changed: 26 FEB 2020

# Set flags for tasks:

#genderToDo <- "female"
genderToDo <- "male"
doFitting <- TRUE
doIndivPlots <- TRUE
doPopulPlots <- TRUE
doContrastPlots <- TRUE

# Set number of mean field variational Bayes iterations:

maxIterVal <- 100

# Set progress reporting flag value:

quietVal <- TRUE

# Set flag for use of colours versus black and white:

colourVersion <- TRUE

# Set flag for subsetting of the individual plots:

doIndivSubset <- FALSE
subsetSize <- 25
layoutVecSubset <- c(5,5)

# Set graphics parameters:

cex.axisVal <- 1.1
cex.labVal <- 1.1

# Load required packages and functions:

library(lattice) ; library(nlme)  ;  library(Matrix)
source("ZOSull.r") 
source("matrixSqrt.r")
source("SolveTwoLevelSparseLeastSquares.r")
source("twoLevGrpSpecCurvContrMFVB.r")

# Read in the data:

growthIndiana <- read.table("growthIndiana.txt",header=TRUE)

if (genderToDo=="female") gender <- 0
if (genderToDo=="male") gender <- 1

# Extract data for male/female subjects:

growthIndForAna <- growthIndiana[growthIndiana$male==gender,]
ageOrig <- growthIndForAna$age
heightOrig <- growthIndForAna$height
idnumRaw <- growthIndForAna$idnum
black <- growthIndForAna$black
idnum <- match(idnumRaw,unique(idnumRaw))
growthIndForAna$idnum <- idnum
iB <- black
iA <- 1-iB

# Create transformed versions of height and age for Bayesian analysis:

meanAge <- mean(ageOrig) ; sdAge <- sd(ageOrig)
meanHeight <- mean(heightOrig) ; sdHeight <- sd(heightOrig)
age <- (ageOrig - meanAge)/sdAge
height <- (heightOrig - meanHeight)/sdHeight

uqID <- unique(idnum)
reBlockInds <- vector("list",length(uqID))

nObs <- length(height)
for (i in 1:length(uqID))
   reBlockInds[[i]] <- (1:nObs)[idnum==i]

# Set up X and Z matrix for mean curve:

numObs <- length(height)
numSubj <- length(unique(idnum))
uqID <- unique(idnum)

Xgbl <- cbind(rep(1,numObs), age, 1 - iA, (1 - iA)*age)
ncXgbl <- ncol(Xgbl)
Xgrp <- cbind(rep(1,numObs), age)
ncXgrp <- ncol(Xgrp)

numIntKnots <- 20
intKnots <- quantile(unique(age),
                     seq(0,1,length=numIntKnots+2))[-c(1,numIntKnots+2)]

range.ageOrig <- c(5.2,20)
range.age <- (range.ageOrig - meanAge)/sdAge

Zgbl <- ZOSull(age, range.age, intKnots)
Kgbl <- ncol(Zgbl)

ZgblA <- iA*Zgbl
ZgblB <- (1 - iA)*Zgbl

Zgbl <- cbind(ZgblA, ZgblB)
ncZgbl <- ncol(Zgbl)

# Set up Z matrix for subject specific curves:

numIntKnotsGrp <- 10
intKnotsGrp <- quantile(unique(age),
        seq(0,1,length=numIntKnotsGrp+2))[-c(1,numIntKnotsGrp+2)]

Zgrp <- ZOSull(age, range.age, intKnotsGrp)
Kgrp <- ncol(Zgrp)
ncZgrp <- ncol(Zgrp)

# Do setup for plotting:

ng <- 101
agegOrig <- seq(min(ageOrig),max(ageOrig),length=ng)
ageg <- (agegOrig - meanAge)/sdAge
Xg <- cbind(rep(1,ng),ageg)
Zg <- ZOSull(ageg,range.age,intKnots)
ZGrpg <- ZOSull(ageg,range.age,intKnotsGrp)

# Specify hyperparameters:

mu.beta <- rep(0, ncXgbl)
Sigma.beta <- 1e10 * diag(ncXgbl)
s.eps <- 1e5
nu.eps <- 1
s.gbl.A <- 1e5
nu.gbl.A <- 1
s.gbl.B <- 1e5
nu.gbl.B <- 1
s.Sigma <- rep(1e5, 2)
nu.Sigma <- 2
s.grp <- 1e5
nu.grp <- 1

m <- length(unique(idnum))

nVec <- table(idnum)
sumnVec <- sum(nVec)

x <- age
y <- height

ncZgbl <- ncol(Zgbl)
ncZgrp <- ncol(Zgrp)

if (doFitting)  
{
   # Set up data lists:

   yList <- vector("list",m)
   XgblList <- vector("list",m)
   XgrpList <- vector("list",m)
   ZgblList <- vector("list",m)
   ZgrpList <- vector("list",m)

   for (i in 1:m) 
   {
      yList[[i]] <- y[idnum==i]
      XgblList[[i]] <- Xgbl[idnum==i,]
      XgrpList[[i]] <- Xgrp[idnum==i,]
      ZgblList[[i]] <- Zgbl[idnum==i,]
      ZgrpList[[i]] <- Zgrp[idnum==i,]
   }

   # Call twoLevGrpSpecCurvContrMFVB():

   twoLevGrpSpecCurvContrMFVBObj <-twoLevGrpSpecCurvContrMFVB(yList,XgblList,XgrpList,ZgblList,ZgrpList,
                                   mu.beta,Sigma.beta,s.eps,nu.eps,s.gbl.A,nu.gbl.A,s.gbl.B,nu.gbl.B,
                                   s.Sigma,nu.Sigma,s.grp,nu.grp,Kgbl,Kgrp,maxIter=maxIterVal,quiet=quietVal)

   # Unpack the twoLevGrpSpecCurvContrMFVB() fit object:

   mu.q.betauGbl <- twoLevGrpSpecCurvContrMFVBObj$mu.q.betauGbl
   mu.q.uLinGrp <- twoLevGrpSpecCurvContrMFVBObj$mu.q.uLinGrp
   Sigma.q.betauGbl <- twoLevGrpSpecCurvContrMFVBObj$Sigma.q.betauGbl
   Sigma.q.uLinGrp <- twoLevGrpSpecCurvContrMFVBObj$Sigma.q.uLinGrp
   E.q.betauGblLinGrp <- twoLevGrpSpecCurvContrMFVBObj$E.q.betauGblLinGrp

   mu.q.beta.A <- mu.q.betauGbl[1:(ncXgbl/2)]
   mu.q.beta.BvsA <- mu.q.betauGbl[(ncXgbl/2 + 1):ncXgbl]
   mu.q.uGbl.A <- mu.q.betauGbl[(ncXgbl+1):(ncXgbl+Kgbl)]
   mu.q.uGbl.B <- mu.q.betauGbl[(ncXgbl+Kgbl+1):(ncXgbl+2*Kgbl)]

   mu.q.uGrpLin <- vector("list", m)  ;  mu.q.uGrpSpl <- vector("list", m)

   for (i in 1:m) 
   {
      mu.q.uGrpLin[[i]] <- mu.q.uLinGrp[[i]][1:ncXgrp]
      mu.q.uGrpSpl[[i]] <- mu.q.uLinGrp[[i]][(ncXgrp+1):(ncXgrp+Kgrp)]
   }
}

if (doIndivPlots|doPopulPlots)
{
   if (colourVersion)
      colourVec <- c("dodgerblue","deeppink","blue","darkmagenta")

   if (!colourVersion)
      colourVec <- c("grey70","black","grey70","black")
}

if (doIndivPlots)
{
   Cg <- cbind(Xg,Xg,Zg,Zg,Xg,ZGrpg)

   fAgMFVB <- (Xg%*%mu.q.betauGbl[1:(ncXgbl/2)] + Zg%*%mu.q.betauGbl[(ncXgbl+1):(ncXgbl+Kgbl)])
   fBgMFVB <- (cbind(Xg,Xg)%*%mu.q.betauGbl[1:ncXgbl] + Zg%*%mu.q.betauGbl[(ncXgbl+Kgbl+1):(ncXgbl+2*Kgbl)])

   gMFVB <- vector("list",m)
   sdMFVB <- vector("list",m)

   curvEstsMFVB <- matrix(NA,ng,m)
   credLowerStreamMFVB <- matrix(NA,ng,m)
   credUpperStreamMFVB <- matrix(NA,ng,m)
   curvEstsMFVBorig <- matrix(NA,ng,m)
   credLowerStreamMFVBorig <- matrix(NA,ng,m)
   credUpperStreamMFVBorig <- matrix(NA,ng,m)

   for (i in 1:m)
   {
      indsi <- (1:numObs)[idnum==i]

      gMFVB[[i]] <- cbind(Xg,ZGrpg)%*%c(mu.q.uGrpLin[[i]],mu.q.uGrpSpl[[i]])
      curvEstsMFVB[,i] <- (iA[indsi[1]]*fAgMFVB + iB[indsi[1]]*fBgMFVB + gMFVB[[i]])

      indsGbl <- 1:(ncXgbl+ncZgbl)
      lenGbl <- ncXgbl+ncZgbl
      lenGrp <- ncXgrp+ncZgrp

      Sigma.q.betauStream <- Sigma.q.betauGbl
      Sigma.q.betauStream <- magic::adiag(Sigma.q.betauStream, Sigma.q.uLinGrp[[i]])
      Sigma.q.betauStream[indsGbl,(lenGbl+1):(lenGbl+lenGrp)] <- E.q.betauGblLinGrp[[i]]
      Sigma.q.betauStream[(lenGbl+1):(lenGbl+lenGrp),indsGbl] <- t(E.q.betauGblLinGrp[[i]])

      sdMFVB[[i]] <- sqrt(diag(Cg%*%Sigma.q.betauStream%*%t(Cg)))

      credLowerStreamMFVB[,i] <- curvEstsMFVB[,i] - qnorm(0.995)*sdMFVB[[i]]
      credUpperStreamMFVB[,i] <- curvEstsMFVB[,i] + qnorm(0.995)*sdMFVB[[i]]

      # Back transform curve estimates to original scale:

      curvEstsMFVBorig[,i] <- meanHeight + sdHeight*curvEstsMFVB[,i]
      credLowerStreamMFVBorig[,i] <- meanHeight + sdHeight*credLowerStreamMFVB[,i]
      credUpperStreamMFVBorig[,i] <- meanHeight + sdHeight*credUpperStreamMFVB[,i]
   }

   if (gender == 0)
   {
      layoutVec <- c(10,10)
      keyText <- list(c("black females","white females"),cex=1.85)
      ylimVal <- c(90,190)
   }  

   if (gender == 1)
   {
      layoutVec <- c(9,13)
      keyText <- list(c("black males","white males"),cex=1.55)
      ylimVal <- c(90,220)
   } 

   if (!doIndivSubset)
   {
      if (gender==0) layoutVec <- c(10,10)
      if (gender==1) layoutVec <- c(9,13)
   }
   if (doIndivSubset)
   {
      layoutVec <- layoutVecSubset
   }

   if (!doIndivSubset)
   {
      growthIndForLattice <- growthIndForAna
      curvEstsMFVBforLattice <- curvEstsMFVBorig
      credLowerStreamMFVBforLattice <- credLowerStreamMFVBorig
      credUpperStreamMFVBforLattice <- credUpperStreamMFVBorig
   }

   if (doIndivSubset)
   {
      # Obtain version of growthIndForLattice 
      # with subsetting:

      uqIdnum <- unique(growthIndForAna$idnum)
      if (genderToDo=="male") set.seed(4)
      if (genderToDo=="female") set.seed(3)
      idnumSub <- sort(sample(uqIdnum,subsetSize,replace=FALSE))

      # Determine rows of growthInd to be used in the
      # subset:

      idnumOrig <- growthIndForAna$idnum

      indsToKeep <- NULL
      for (i in 1:length(idnumOrig))
      {
         if (any(idnumOrig[i]==idnumSub)) indsToKeep <- c(indsToKeep,i)
      }

      growthIndForLattice <- growthIndForAna[indsToKeep,]

      # Organise mapping from old retained ID numbers to new ones.

      idnumSubset <- growthIndForLattice$idnum
      uqIdnumOld <- as.numeric(unique(idnumSubset))

      idnumNew <- match(idnumSubset,unique(idnumSubset))

      growthIndForLattice$idnum <- idnumNew

      credLowerStreamMFVBforLattice <- NULL
      credUpperStreamMFVBforLattice <- NULL
      curvEstsMFVBforLattice <- NULL
      for (id in 1:length(unique(idnumNew)))
      {
         curvEstsMFVBforLattice <- cbind(curvEstsMFVBforLattice,curvEstsMFVBorig[,uqIdnumOld[id]])
      }
   }

   figFit <- xyplot(height~age|idnum,groups=idnum,data=growthIndForLattice,
                   layout=layoutVec,ylim=ylimVal, 
                   xlab=list("age (years)",cex=1.5),
    	           ylab=list("height (centimetres)",cex=1.5),
                   scales=list(cex=1.25),strip=FALSE,as.table=TRUE,
                   key = list(title="",
                              columns = 2, 
                              lines = list(lty = rep(1,2),col = colourVec[2:1],lwd = rep(2,2),pch=rep(1,2)),
                              text = keyText),
                   panel=function(x,y,subscripts,groups)
                  {
                     panel.grid()

                     columnNum <- panel.number()
                     colourInd <- black[subscripts[1]] + 1
 
                     panel.polygon(c(agegOrig,rev(agegOrig)),
                                   c(credLowerStreamMFVBforLattice[,columnNum],
                                     rev(credUpperStreamMFVBforLattice[,columnNum])),
                                     col=colourVec[colourInd+4],border=FALSE)
 
                     panel.xyplot(agegOrig,curvEstsMFVBforLattice[,columnNum],type="l",
                                   col=colourVec[colourInd+2]) 

                     panel.superpose(x,y,subscripts,groups,
                                    type="b",col=colourVec[colourInd],pch=1,cex=0.5)

                  })
       
   print(figFit)

   readline("Hit Enter to continue.\n")
}

if (doPopulPlots)
{
   CgA <- cbind(Xg,Zg)
   indsA <- c(1:(ncXgbl/2),(ncXgbl+1):(ncXgbl+Kgbl))
   fAgMFVB <- as.vector(CgA%*%mu.q.betauGbl[indsA])
   sdAgMFVB <- as.vector(sqrt(diag(CgA%*%Sigma.q.betauGbl[indsA,indsA]%*%t(CgA))))
   lowAgMFVB <- fAgMFVB - qnorm(0.975)*sdAgMFVB
   uppAgMFVB <- fAgMFVB + qnorm(0.975)*sdAgMFVB

   CgB <- cbind(Xg,Xg,Zg)
   indsB <- c(1:ncXgbl,(ncXgbl+Kgbl+1):(ncXgbl+2*Kgbl))
   fBgMFVB <- as.vector(CgB%*%mu.q.betauGbl[indsB])
   sdBgMFVB <- as.vector(sqrt(diag(CgB%*%Sigma.q.betauGbl[indsB,indsB]%*%t(CgB))))
   lowBgMFVB <- fBgMFVB - qnorm(0.975)*sdBgMFVB
   uppBgMFVB <- fBgMFVB + qnorm(0.975)*sdBgMFVB

   # Back transform curve estimates to original scale:

   fAgMFVBorig <- meanHeight + sdHeight*fAgMFVB
   lowAgMFVBorig <- meanHeight + sdHeight*lowAgMFVB
   uppAgMFVBorig <- meanHeight + sdHeight*uppAgMFVB

   fBgMFVBorig <- meanHeight + sdHeight*fBgMFVB
   lowBgMFVBorig <- meanHeight + sdHeight*lowBgMFVB
   uppBgMFVBorig <- meanHeight + sdHeight*uppBgMFVB

   lwdVal <- 2
   plot(0,type="n",bty="l",xlim=range(agegOrig),
        ylim=range(c(lowAgMFVBorig,uppAgMFVBorig,lowBgMFVBorig,uppBgMFVBorig)),
        xlab="age (years)",ylab="height (centimetres)",
        cex.axis=cex.axisVal,cex.lab=cex.labVal)
   lines(agegOrig,fAgMFVBorig,col=colourVec[3],lwd=lwdVal)
   lines(agegOrig,lowAgMFVBorig,col=colourVec[3],lty=2,lwd=lwdVal)
   lines(agegOrig,uppAgMFVBorig,col=colourVec[3],lty=2,lwd=lwdVal)

   lines(agegOrig,fBgMFVBorig,col=colourVec[4],lwd=lwdVal)
   lines(agegOrig,lowBgMFVBorig,col=colourVec[4],lty=2,lwd=lwdVal)
   lines(agegOrig,uppBgMFVBorig,col=colourVec[4],lty=2,lwd=lwdVal)

   # Add legend:

   legendCexVal <- 1.65
   if (gender==0) legendString <- c("black females","white females")
   if (gender==1) legendString <- c("black males","white males")
   legend("topleft",legend=legendString,lty=rep(1,2),lwd=rep(lwdVal,2),
          col=colourVec[4:3],cex=legendCexVal)

   readline("Hit Enter to continue.\n")
}

if (doContrastPlots)
{
   if (colourVersion)
   {
      lineCol <- "darkgreen"
      shadeCol <- "palegreen"
      axisCol <- "slateblue"
   }
   if (!colourVersion)
   {
      lineCol <- "black"
      shadeCol <- "black"
      axisCol <- "black"
   }

   if (gender==0) fileNameCurr <- "growthIndContrFem.pdf"
   if (gender==1) fileNameCurr <- "growthIndContrMal.pdf"

   CgA <- cbind(Xg,Zg)
   CgB <- cbind(Xg,Xg,Zg)

   Contg <- cbind(Xg,-Zg,Zg)

   coefsA <- c(mu.q.beta.A, mu.q.uGbl.A)
   coefsBvsA <- c(mu.q.beta.A, mu.q.beta.BvsA, mu.q.uGbl.B)

   fAMFVB <- as.vector(CgA%*%coefsA)
   fBMFVB <- as.vector(CgB%*%coefsBvsA)

   sdMFVB <- sqrt(diag(Contg%*%Sigma.q.betauGbl[((ncXgbl/2)+1):(ncXgbl+ncZgbl),
                  ((ncXgbl/2)+1):(ncXgbl+ncZgbl)]%*%t(Contg)))

   ContgMFVB <- fBMFVB - fAMFVB
   ContgMFVBLower <- ContgMFVB - qnorm(0.975)*sdMFVB
   ContgMFVBUpper <- ContgMFVB + qnorm(0.975)*sdMFVB

   # Back-transform contrast curve to original units.

   ContgMFVBOrig <- ContgMFVB * sdHeight
   ContgLowerOrig <- ContgMFVBLower * sdHeight
   ContgUpperOrig <- ContgMFVBUpper * sdHeight

   ylimVal <- range(c(ContgLowerOrig,ContgUpperOrig))

   plot(agegOrig,ContgMFVBOrig,bty="l",type="n",xlab="age (years)",
        ylab="mean height difference (centimetres)",ylim=ylimVal,
        cex.axis=cex.axisVal,cex.lab=cex.labVal)

   lines(agegOrig,ContgLowerOrig,col=lineCol,lty=2,lwd=2)
   lines(agegOrig,ContgUpperOrig,col=lineCol,lty=2,lwd=2)

   lines(agegOrig,ContgMFVBOrig,col=lineCol,lwd=2)
   abline(0,0,col=axisCol)
}

############ End of growthIndianaExample ###########
