########### R function: threeLevGrpSpecCurvMFVB ###########

# For performing a streamlined variational approximate Bayesian 
# three-level group specific curve data.

# Last changed: 26 FEB 2020

threeLevGrpSpecCurvMFVB <- function(yList,XList,ZgblList,Zgrp1List,Zgrp2List,nVec,oMat,
                                    mu.beta,Sigma.beta,s.eps,nu.eps,s.gbl,nu.gbl,s.SigmaL1,
                                    nu.SigmaL1,s.SigmaL2,nu.SigmaL2,s.grp1,nu.grp1,s.grp2,
                                    nu.grp2,maxIter=200,tolerance=1e-5,plotLogML=TRUE,quiet=FALSE) 
{
   # Obtain dimension variables:

   m <- length(yList)
   ncX <- 2
   q1 <- ncX  ;  q2 <- ncX

   Kgbl <- ncol(ZgblList[[1]][[1]])
   Kgrp1 <- ncol(Zgrp1List[[1]][[1]])
   Kgrp2 <- ncol(Zgrp2List[[1]][[1]])

   # Obtain the Cgbl and Cgrp lists:

   CgblList <- vector("list",m)
   Cgrp1List <- vector("list",m)
   Cgrp2List <- vector("list",m)

   for (i in 1:m) 
   {
      CgblList[[i]] <- vector("list",nVec[i])
      Cgrp1List[[i]] <- vector("list",nVec[i])
      Cgrp2List[[i]] <- vector("list",nVec[i])
   }

   for (i in 1:m) 
   {
      for (j in 1:nVec[i]) 
      {
         CgblList[[i]][[j]] <- cbind(XList[[i]][[j]], ZgblList[[i]][[j]])
         Cgrp1List[[i]][[j]] <- cbind(XList[[i]][[j]], Zgrp1List[[i]][[j]])
         Cgrp2List[[i]][[j]] <- cbind(XList[[i]][[j]], Zgrp2List[[i]][[j]])
      }
   }

   # Allocate space for the b, B, Bdot and Bdotdot lists and
   # allocate lists for q-density parameters:

   b <- vector("list",m)
   B <- vector("list",m)
   Bdot <- vector("list",m)
   Bdotdot <- vector("list",m)

   A22L1 <- vector("list",m)
   A22L2 <- vector("list",m)
   A12L2comma <- vector("list",m)

   mu.q.uLinGrp1 <- vector("list",m)  ;  mu.q.uLinGrp2 <- vector("list",m)
   Sigma.q.uLinGrp1 <- vector("list",m)  ;  Sigma.q.uLinGrp2 <- vector("list",m)
   E.q.betauGblLinGrp1 <- vector("list",m)  ;  E.q.betauGblLinGrp2 <- vector("list",m)
   E.q.uLinGrp1uLinGrp2 <- vector("list",m)

   for (i in 1:m) 
   {
      b[[i]] <- vector("list",nVec[i])
      B[[i]] <- vector("list",nVec[i])
      Bdot[[i]] <- vector("list",nVec[i])
      Bdotdot[[i]] <- vector("list",nVec[i])

      A22L2[[i]] <- vector("list",nVec[i])
      A12L2comma[[i]] <- vector("list",nVec[i])

      mu.q.uLinGrp2[[i]] <- vector("list", nVec[i])
      Sigma.q.uLinGrp2[[i]] <- vector("list", nVec[i])
      E.q.betauGblLinGrp2[[i]] <- vector("list", nVec[i])
      E.q.uLinGrp1uLinGrp2[[i]] <- vector("list", nVec[i])
   }

   # Do MFVB iterations:

   mu.q.recip.sigsq.eps <- 1
   mu.q.recip.a.eps <- 1
   mu.q.recip.sigsq.gbl <- 1
   mu.q.recip.a.gbl <- 1

   M.q.inv.SigmaL1 <- diag(ncX)
   M.q.inv.SigmaL2 <- diag(ncX)
   M.q.inv.A.SigmaL1 <- diag(q1)
   M.q.inv.A.SigmaL2 <- diag(q2)
   mu.q.recip.sigsq.grp1 <- 1
   mu.q.recip.sigsq.grp2 <- 1
   mu.q.recip.a.grp1 <- 1
   mu.q.recip.a.grp2 <- 1

   xi.q.sigsq.eps <- nu.eps + sum(oMat)
   xi.q.a.sigsq.eps <- nu.eps + 1
   xi.q.sigsq.gbl <- nu.gbl + Kgbl
   xi.q.a.sigsq.gbl <- nu.gbl + 1
   xi.q.SigmaL1 <- nu.SigmaL1 + (2 * q1) - 2 + m
   xi.q.A.SigmaL1 <- nu.SigmaL1 + q1
   xi.q.SigmaL2 <- nu.SigmaL2 + (2 * q2) - 2 + m
   xi.q.A.SigmaL2 <- nu.SigmaL2 + q2
   xi.q.sigsq.grp1 <- nu.grp1 + m*Kgrp1
   xi.q.a.sigsq.grp1 <- nu.grp1 + 1
   xi.q.sigsq.grp2 <- nu.grp2 + sum(nVec)*Kgrp2
   xi.q.a.sigsq.grp2 <- nu.grp2 + 1

   Lambda.q.A.SigmaL1 <- matrix(0, q1, q1)
   Lambda.q.A.SigmaL2 <- matrix(0, q2, q2)

   converged <- FALSE
   itnum <- 0
   logMLcurr <- -Inf
   logMLgrid <- NULL
   while (!converged) 
   {
      itnum <- itnum + 1
      if (!quiet) cat("Performing iteration number ",itnum,".\n",sep = "")

      # Set current log(ML) value to previous one:

      logMLprev <- logMLcurr

      for (i in 1:m) 
      {
         A22L1[[i]] <- matrix(0,(ncX+Kgrp1),(ncX+Kgrp1))
         for (j in 1:nVec[i]) 
         {

            # Obtain b,B and Bdot lists:

            b[[i]][[j]] <- c((yList[[i]][[j]] * sqrt(mu.q.recip.sigsq.eps)),
                                 matrixSqrt(solve(sum(nVec)*Sigma.beta)) %*% mu.beta,
                                 rep(0,Kgbl),
                                 rep(0,ncX),
                                 rep(0,Kgrp1),
                                 rep(0,ncX),
                                 rep(0,Kgrp2))

            B[[i]][[j]] <- rbind(cbind((XList[[i]][[j]] * sqrt(mu.q.recip.sigsq.eps)), 
                                        (ZgblList[[i]][[j]] * sqrt(mu.q.recip.sigsq.eps))),
                               cbind(matrixSqrt(solve(sum(nVec) * Sigma.beta)), matrix(0, ncX, Kgbl)),
                               cbind(matrix(0, Kgbl, ncX), (diag(Kgbl)/sqrt(sum(nVec)))*sqrt(mu.q.recip.sigsq.gbl)),
                               matrix(0, ncX, ncX + Kgbl),
                               matrix(0, Kgrp1, ncX + Kgbl),
                               matrix(0, ncX, ncX + Kgbl),
                               matrix(0, Kgrp2, ncX + Kgbl))

            Bdot[[i]][[j]] <- rbind(cbind((XList[[i]][[j]] * sqrt(mu.q.recip.sigsq.eps)), 
                                     (Zgrp1List[[i]][[j]] * sqrt(mu.q.recip.sigsq.eps))),
                                        matrix(0, ncX, ncX + Kgrp1),
                                        matrix(0, Kgbl, ncX + Kgrp1),
                                        cbind(matrixSqrt((1/nVec[i]) * M.q.inv.SigmaL1), matrix(0, q1, Kgrp1)),
                                        cbind(matrix(0, Kgrp1, ncX), diag(Kgrp1)*sqrt((1/nVec[i])*mu.q.recip.sigsq.grp1)),
                                        matrix(0, ncX, ncX + Kgrp1),
                                        matrix(0, Kgrp2, ncX + Kgrp1))

            Bdotdot[[i]][[j]] <- rbind(cbind((XList[[i]][[j]] * sqrt(mu.q.recip.sigsq.eps)), 
                                           (Zgrp2List[[i]][[j]] * sqrt(mu.q.recip.sigsq.eps))),
                                           matrix(0, ncX, ncX + Kgrp2),
                                           matrix(0, Kgbl, ncX + Kgrp2),
                                           matrix(0, ncX, ncX + Kgrp2),
                                           matrix(0, Kgrp1, ncX + Kgrp2),
                                           cbind(matrixSqrt(M.q.inv.SigmaL2), matrix(0, q2, Kgrp2)),
                                           cbind(matrix(0,Kgrp2,ncX), diag(Kgrp2)*sqrt(mu.q.recip.sigsq.grp2)))

            A22L1[[i]] <- A22L1[[i]] + crossprod(Bdot[[i]][[j]])
            A22L2[[i]][[j]] <- crossprod(Bdotdot[[i]][[j]])
            A12L2comma[[i]][[j]] <- crossprod(Bdot[[i]][[j]], Bdotdot[[i]][[j]])

         }
      }

      # Call the SolveThreeLevelSparseLeastSquares algorithm:

      S1 <- SolveThreeLevelSparseLeastSquares(b, B, Bdot, Bdotdot)

      # Extract coefficients estimates and covariance matrix sub-blocks:

      mu.q.betauGbl <- S1$x1
      Sigma.q.betauGbl <- S1$Asup11

      for (i in 1:m) 
      {
         mu.q.uLinGrp1[[i]] <- S1$x2L1[[i]]  ;  Sigma.q.uLinGrp1[[i]] <- S1$Asup22L1[[i]]
         E.q.betauGblLinGrp1[[i]] <- S1$Asup12L1[[i]]
         for (j in 1:nVec[i]) 
         {
            mu.q.uLinGrp2[[i]][[j]] <- S1$x2L2[[i]][[j]]  ;   Sigma.q.uLinGrp2[[i]][[j]] <- S1$Asup22L2[[i]][[j]]
            E.q.betauGblLinGrp2[[i]][[j]] <- S1$Asup12L2[[i]][[j]]
            E.q.uLinGrp1uLinGrp2[[i]][[j]] <- S1$Asup1comma2L2[[i]][[j]]
         }
      }

      # Update lambda.q.sigsq.gbl:

      lambda.q.sigsq.gbl <- (mu.q.recip.a.gbl + sum(mu.q.betauGbl[3:(2+Kgbl)]^2)
                               + sum(diag(Sigma.q.betauGbl[3:(2+Kgbl),3:(2+Kgbl)])))

      # Update lambda.q.sigsq.eps, Lambda.q.SigmaL1, Lambda.q.SigmaL1,
      # lambda.q.sigsq.grp1 and lambda.q.sigsq.grp2:

      Lambda.q.SigmaL1 <- M.q.inv.A.SigmaL1
      Lambda.q.SigmaL2 <- M.q.inv.A.SigmaL2

      lambda.q.sigsq.eps <- mu.q.recip.a.eps
      lambda.q.sigsq.grp1 <- mu.q.recip.a.grp1
      lambda.q.sigsq.grp2 <- mu.q.recip.a.grp2

      for (i in 1:m) 
      {
         Lambda.q.SigmaL1 <- Lambda.q.SigmaL1 + tcrossprod(mu.q.uLinGrp1[[i]][1:2]) + Sigma.q.uLinGrp1[[i]][1:2,1:2]
         lambda.q.sigsq.grp1 <- (lambda.q.sigsq.grp1 + sum(mu.q.uLinGrp1[[i]][3:(2+Kgrp1)]^2)
                                       + sum(diag(Sigma.q.uLinGrp1[[i]][3:(2+Kgrp1),3:(2+Kgrp1)])))

         for (j in 1:nVec[i]) 
         {
            residVec <- (yList[[i]][[j]] - as.vector(CgblList[[i]][[j]]%*%mu.q.betauGbl)
                           - as.vector(Cgrp1List[[i]][[j]]%*%mu.q.uLinGrp1[[i]])
                           - as.vector(Cgrp2List[[i]][[j]]%*%mu.q.uLinGrp2[[i]][[j]]))
            lambda.q.sigsq.eps <- lambda.q.sigsq.eps + sum(residVec^2)
            lambda.q.sigsq.eps <- lambda.q.sigsq.eps + sum(diag(tcrossprod(
                                     crossprod(CgblList[[i]][[j]]),Sigma.q.betauGbl)))
            lambda.q.sigsq.eps <- lambda.q.sigsq.eps + sum(diag(tcrossprod(
                                     crossprod(Cgrp1List[[i]][[j]]),Sigma.q.uLinGrp1[[i]])))
            lambda.q.sigsq.eps <- lambda.q.sigsq.eps + sum(diag(tcrossprod(
                                    crossprod(Cgrp2List[[i]][[j]]),Sigma.q.uLinGrp2[[i]][[j]])))
            lambda.q.sigsq.eps <- lambda.q.sigsq.eps + 2*sum(diag(tcrossprod(
                                    crossprod(Cgrp1List[[i]][[j]],CgblList[[i]][[j]]),t(E.q.betauGblLinGrp1[[i]]))))
            lambda.q.sigsq.eps <- lambda.q.sigsq.eps + 2*sum(diag(tcrossprod(
                                    crossprod(Cgrp2List[[i]][[j]],CgblList[[i]][[j]]),
                                    t(E.q.betauGblLinGrp2[[i]][[j]]))))
            lambda.q.sigsq.eps <- lambda.q.sigsq.eps + 2*sum(diag(tcrossprod(
                                    crossprod(Cgrp1List[[i]][[j]],Cgrp2List[[i]][[j]]),
                                    E.q.uLinGrp1uLinGrp2[[i]][[j]])))

            Lambda.q.SigmaL2 <- (Lambda.q.SigmaL2 + tcrossprod(mu.q.uLinGrp2[[i]][[j]][1:2]) 
                                  + Sigma.q.uLinGrp2[[i]][[j]][1:2,1:2])
            lambda.q.sigsq.grp2 <- (lambda.q.sigsq.grp2 + sum(mu.q.uLinGrp2[[i]][[j]][3:(2+Kgrp2)]^2)
                                       + sum(diag(Sigma.q.uLinGrp2[[i]][[j]][3:(2+Kgrp2),3:(2+Kgrp2)])))
         }
      }

      # Do each of the mean reciprocal and inverse updates for variance and covariance parameters:

      M.q.inv.SigmaL1 <- (xi.q.SigmaL1 - q1 + 1) * solve(Lambda.q.SigmaL1)
      M.q.inv.SigmaL2 <- (xi.q.SigmaL2 - q2 + 1) * solve(Lambda.q.SigmaL2)

      mu.q.recip.sigsq.eps <- xi.q.sigsq.eps/lambda.q.sigsq.eps
      mu.q.recip.sigsq.gbl <- xi.q.sigsq.gbl/lambda.q.sigsq.gbl
      mu.q.recip.sigsq.grp1 <- xi.q.sigsq.grp1/lambda.q.sigsq.grp1
      mu.q.recip.sigsq.grp2 <- xi.q.sigsq.grp2/lambda.q.sigsq.grp2

      # Do each of the mean reciprocal and inverse updates for auxiliary parameters:

      lambda.q.a.sigsq.eps <- mu.q.recip.sigsq.eps + (1 / (nu.eps * s.eps^2))
      mu.q.recip.a.eps <- xi.q.a.sigsq.eps / lambda.q.a.sigsq.eps

      lambda.q.a.sigsq.gbl <- mu.q.recip.sigsq.gbl + (1 / (nu.gbl * s.gbl^2))
      mu.q.recip.a.gbl <- xi.q.a.sigsq.gbl / lambda.q.a.sigsq.gbl

      diag(Lambda.q.A.SigmaL1) <- diag(M.q.inv.SigmaL1) +  (1/(nu.SigmaL1 * s.SigmaL1^2))
      M.q.inv.A.SigmaL1 <- xi.q.A.SigmaL1 * solve(Lambda.q.A.SigmaL1)

      diag(Lambda.q.A.SigmaL2) <- diag(M.q.inv.SigmaL2) +  (1/(nu.SigmaL2 * s.SigmaL2^2))
      M.q.inv.A.SigmaL2 <- xi.q.A.SigmaL2 * solve(Lambda.q.A.SigmaL2)

      lambda.q.a.sigsq.grp1 <- mu.q.recip.sigsq.grp1 + (1 / (nu.grp1 * s.grp1^2))
      mu.q.recip.a.grp1 <- xi.q.a.sigsq.grp1 / lambda.q.a.sigsq.grp1

      lambda.q.a.sigsq.grp2 <- mu.q.recip.sigsq.grp2 + (1 / (nu.grp2 * s.grp2^2))
      mu.q.recip.a.grp2 <- xi.q.a.sigsq.grp2 / lambda.q.a.sigsq.grp2

      detTerm1 <- as.numeric(determinant(Sigma.q.betauGbl)$modulus)
      detTerm2 <- 0   ;  detTerm3 <- 0
      for (i in 1:m)
      {
         sumTerm <- 0
         for (j in 1:nVec[i])
         {
            sumTerm <- sumTerm + A12L2comma[[i]][[j]]%*%solve(A22L2[[i]][[j]],t(A12L2comma[[i]][[j]]))
            detTerm3 <- detTerm3 + as.numeric(determinant(A22L2[[i]][[j]])$modulus)
         }
         detTerm2 <- detTerm2 + as.numeric(determinant(A22L1[[i]] - sumTerm)$modulus)
      }
      det.Sigma.q.betau <- detTerm1 - detTerm2 - detTerm3

      # Obtain the current approximate marginal log-likelihood:

      logMLcurr <- threeLevGrpSpecCurvLogML(p=ncX, m=m, nVec=nVec, oMat=oMat, q1=q1, q2=q2, Kgbl=Kgbl, Kgrp1=Kgrp1, Kgrp2=Kgrp2,
            mu.beta=mu.beta, Sigma.beta=Sigma.beta, yList=yList, CgblList=CgblList, Cgrp1List=Cgrp1List, Cgrp2List=Cgrp2List,
            mu.q.beta=mu.q.betauGbl[1:ncX], Sigma.q.beta=Sigma.q.betauGbl[1:ncX,1:ncX], mu.q.betauGbl=mu.q.betauGbl,
            Sigma.q.betauGbl=Sigma.q.betauGbl,mu.q.uLinGrp1=mu.q.uLinGrp1, Sigma.q.uLinGrp1=Sigma.q.uLinGrp1, mu.q.uLinGrp2=mu.q.uLinGrp2,
            Sigma.q.uLinGrp2=Sigma.q.uLinGrp2, E.q.betauGblLinGrp1=E.q.betauGblLinGrp1, E.q.betauGblLinGrp2=E.q.betauGblLinGrp2,
            E.q.uLinGrp1uLinGrp2=E.q.uLinGrp1uLinGrp2, M.q.inv.SigmaL1=M.q.inv.SigmaL1, M.q.inv.SigmaL2=M.q.inv.SigmaL2,
            mu.q.recip.sigsq.gbl=mu.q.recip.sigsq.gbl, mu.q.recip.sigsq.grp1=mu.q.recip.sigsq.grp1,
            mu.q.recip.sigsq.grp2=mu.q.recip.sigsq.grp2, nu.eps=nu.eps, nu.gbl=nu.gbl, nu.grp1=nu.grp1, nu.grp2=nu.grp2,
            nu.SigmaL1=nu.SigmaL1, nu.SigmaL2=nu.SigmaL2, mu.q.recip.sigsq.eps=mu.q.recip.sigsq.eps, mu.q.recip.a.eps=mu.q.recip.a.eps,
            xi.q.sigsq.eps=xi.q.sigsq.eps, lambda.q.sigsq.eps=lambda.q.sigsq.eps, s.eps=s.eps, xi.q.a.sigsq.eps=xi.q.a.sigsq.eps,
            mu.q.recip.a.gbl=mu.q.recip.a.gbl, xi.q.sigsq.gbl=xi.q.sigsq.gbl, lambda.q.a.sigsq.eps=lambda.q.a.sigsq.eps,
            lambda.q.sigsq.gbl=lambda.q.sigsq.gbl, s.gbl=s.gbl, xi.q.a.sigsq.gbl=xi.q.a.sigsq.gbl,
            lambda.q.a.sigsq.gbl=lambda.q.a.sigsq.gbl, mu.q.recip.a.grp1=mu.q.recip.a.grp1, xi.q.sigsq.grp1=xi.q.sigsq.grp1,
            lambda.q.sigsq.grp1=lambda.q.sigsq.grp1, s.grp1=s.grp1, xi.q.a.sigsq.grp1=xi.q.a.sigsq.grp1,
            lambda.q.a.sigsq.grp1=lambda.q.a.sigsq.grp1, mu.q.recip.a.grp2=mu.q.recip.a.grp2, xi.q.sigsq.grp2=xi.q.sigsq.grp2,
            lambda.q.sigsq.grp2=lambda.q.sigsq.grp2, s.grp2=s.grp2, xi.q.a.sigsq.grp2=xi.q.a.sigsq.grp2,
            lambda.q.a.sigsq.grp2=lambda.q.a.sigsq.grp2, M.q.inv.A.SigmaL1=M.q.inv.A.SigmaL1, Lambda.q.SigmaL1=Lambda.q.SigmaL1,
            M.q.inv.A.SigmaL2=M.q.inv.A.SigmaL2, Lambda.q.SigmaL2=Lambda.q.SigmaL2, xi.q.A.SigmaL1=xi.q.A.SigmaL1,
            xi.q.SigmaL1=xi.q.SigmaL1, s.SigmaL1=s.SigmaL1, Lambda.q.A.SigmaL1=Lambda.q.A.SigmaL1, xi.q.A.SigmaL2=xi.q.A.SigmaL2,
            xi.q.SigmaL2=xi.q.SigmaL2, s.SigmaL2=s.SigmaL2, Lambda.q.A.SigmaL2=Lambda.q.A.SigmaL2, det.Sigma.q.betau=det.Sigma.q.betau)

      logMLgrid <- c(logMLgrid,logMLcurr)
      relErr <- abs((logMLcurr/logMLprev) - 1)

      if (!quiet)
         cat("The tolerance is ",tolerance," and the current relative error is ",relErr,".\n",sep = "")

      # Check convergence conditions:

  
      if (relErr<tolerance) converged <- TRUE
      if (itnum==maxIter) 
      {
         converged <- TRUE
         warning("Maximum number of iterations exceeded before convergence achieved.\n")
      }
   }

   if (plotLogML) 
   {
      # Plot the marginal log-likelihood curve:

      logMLgrid <- logMLgrid[10:itnum]
      par(mfrow = c(1,1))
      plot(10:itnum,logMLgrid,ylim = range(logMLgrid),col ="dodgerblue",bty="l",
           xlab = "iteration number",ylab = "approximate marginal log-likelihood")

      readline("Hit Enter to continue.\n")  
   }

   # Return output list:

   return(list(mu.q.betauGbl = mu.q.betauGbl, mu.q.uLinGrp1 = mu.q.uLinGrp1, mu.q.uLinGrp2 = mu.q.uLinGrp2,
               Sigma.q.betauGbl = Sigma.q.betauGbl, Sigma.q.uLinGrp1 = Sigma.q.uLinGrp1,
               E.q.betauGblLinGrp1 = E.q.betauGblLinGrp1,E.q.betauGblLinGrp2 = E.q.betauGblLinGrp2,
               E.q.uLinGrp1uLinGrp2 = E.q.uLinGrp1uLinGrp2, Sigma.q.uLinGrp2 = Sigma.q.uLinGrp2,
               xi.q.sigsq.eps = xi.q.sigsq.eps,lambda.q.sigsq.eps = lambda.q.sigsq.eps,
               xi.q.sigsq.gbl = xi.q.sigsq.gbl,lambda.q.sigsq.gbl = lambda.q.sigsq.gbl,
               xi.q.SigmaL1 = xi.q.SigmaL1, Lambda.q.SigmaL1 = Lambda.q.SigmaL1,
               xi.q.SigmaL2 = xi.q.SigmaL2, Lambda.q.SigmaL2 = Lambda.q.SigmaL2,
               xi.q.sigsq.grp1 = xi.q.sigsq.grp1, lambda.q.sigsq.grp1 = lambda.q.sigsq.grp1,
               xi.q.sigsq.grp2 = xi.q.sigsq.grp2, lambda.q.sigsq.grp2 = lambda.q.sigsq.grp2,
               xi.q.A.SigmaL1 = xi.q.A.SigmaL1, Lambda.q.A.SigmaL1 = Lambda.q.A.SigmaL1,
               xi.q.A.SigmaL2 = xi.q.A.SigmaL2, Lambda.q.A.SigmaL2 = Lambda.q.A.SigmaL2))
}

############ End of threeLevGrpSpecCurvMFVB ############
