########### R function: twoLevGrpSpecCurvContrMFVB ###########

# For performing a streamlined variational approximate Bayesian contrast 
# curve analysis for two-level group specific curve data.

# Last changed: 26 FEB 2020

twoLevGrpSpecCurvContrMFVB <- function(yList,XgblList,XgrpList,ZgblList,ZgrpList,mu.beta,Sigma.beta,
                                       s.eps,nu.eps,s.gbl.A,nu.gbl.A,s.gbl.B,nu.gbl.B,s.Sigma,nu.Sigma,
                                       s.grp,nu.grp,Kgbl,Kgrp,maxIter = 200,quiet = FALSE) 
{
   # Obtain dimension variables:

   m <- length(yList)
   p <- ncol(XgblList[[1]])
   q <- ncol(XgrpList[[1]])
   ncZgbl <- ncol(ZgblList[[1]])
   ncZgrp <- ncol(ZgrpList[[1]])

   # Obtain the Cgbl and Cgrp lists:

   CgblList <- vector("list",m)
   CgrpList <- vector("list",m)
   for (i in 1:m) 
   {
      CgblList[[i]] <- cbind(XgblList[[i]],ZgblList[[i]])
      CgrpList[[i]] <- cbind(XgrpList[[i]],ZgrpList[[i]])
   }

   # Allocate space for the b,B and Bdot lists:

   b <- vector("list",m)
   B <- vector("list",m)
   Bdot <- vector("list",m)

   # Do MFVB iterations:

   mu.q.recip.sigsq.eps <- 1
   mu.q.recip.a.eps <- 1
   mu.q.recip.sigsq.gbl.A <- 1
   mu.q.recip.sigsq.gbl.B <- 1
   mu.q.recip.a.gbl.A <- 1
   mu.q.recip.a.gbl.B <- 1
   M.q.inv.Sigma <- diag(q)
   M.q.inv.A.Sigma <- diag(q)
   mu.q.recip.sigsq.grp <- 1
   mu.q.recip.a.grp <- 1

   xi.q.sigsq.eps <- nu.eps + sum(nVec)
   xi.q.a.sigsq.eps <- nu.eps + 1
   xi.q.sigsq.gbl.A <- nu.gbl.A + Kgbl
   xi.q.a.sigsq.gbl.A <- nu.gbl.A + 1
   xi.q.sigsq.gbl.B <- nu.gbl.B + Kgbl
   xi.q.a.sigsq.gbl.B <- nu.gbl.B + 1
   xi.q.Sigma <- nu.Sigma + (2 * q) - 2 + m
   xi.q.A.Sigma <- nu.Sigma + q
   xi.q.sigsq.grp <- nu.grp + m*Kgrp
   xi.q.a.sigsq.grp <- nu.grp + 1

   Lambda.q.A.Sigma <- matrix(0, q, q)

   itnum <- 0  ;   
   converged <- FALSE
   logMLcurr <- -Inf
   logMLgrid <- NULL
   while (!converged) 
   {
      itnum <- itnum + 1
      if (!quiet)
         cat("Performing iteration number ",itnum," out of ", maxIter,".\n",sep="")

      # create list of matrices:

      mu.q.uLinGrp <- vector("list",m)
      Sigma.q.uLinGrp <- vector("list",m)
      E.q.betauGblLinGrp <- vector("list",m)

      for (i in 1:m) 
      {

         # Obtain b,B and Bdot lists:

         b[[i]] <- c((yList[[i]]*sqrt(mu.q.recip.sigsq.eps)),rep(0,(p + ncZgbl + q + ncZgrp)))

         B[[i]] <- rbind(cbind(XgblList[[i]]*sqrt(mu.q.recip.sigsq.eps), ZgblList[[i]]*sqrt(mu.q.recip.sigsq.eps)),
                         cbind(matrixSqrt(solve(m * Sigma.beta)), matrix(0,p,ncZgbl)),
                         cbind(matrix(0,ncZgbl,p), m^(-1/2) * magic::adiag(diag(Kgbl)*sqrt(mu.q.recip.sigsq.gbl.A), 
                               diag(Kgbl)*sqrt(mu.q.recip.sigsq.gbl.B)) ),
                         matrix(0,q,p+ncZgbl),
                         matrix(0,ncZgrp,(p+ncZgbl)))

         Bdot[[i]] <- rbind(cbind((XgrpList[[i]]*sqrt(mu.q.recip.sigsq.eps)),(ZgrpList[[i]]*sqrt(mu.q.recip.sigsq.eps))),
                            matrix(0,p,(q+ncZgrp)),
                            matrix(0,ncZgbl,q+ncZgrp),
                            cbind(matrixSqrt(M.q.inv.Sigma), matrix(0,q,ncZgrp)),
                            cbind(matrix(0,ncZgrp,q), diag(ncZgrp)*sqrt(mu.q.recip.sigsq.grp)))
      }

      # Call the SolveTwoLevelSparseLeastSquares algorithm:

      S1 <- SolveTwoLevelSparseLeastSquares(b,B,Bdot)

      # Extract coefficients estimates:

      mu.q.betauGbl <- S1$x1

      mu.q.beta <- mu.q.betauGbl[1:p]
      mu.q.uGblA <- mu.q.betauGbl[(p+1):(p+Kgbl)]
      mu.q.uGblB <- mu.q.betauGbl[(p+Kgbl+1):(p+2*Kgbl)]

      # Extract covariance matrix sub-blocks:

      Sigma.q.betauGbl <- S1$Asup11

      Sigma.q.beta <- Sigma.q.betauGbl[1:p, 1:p]
      Sigma.q.uGblA <- Sigma.q.betauGbl[(p+1):(p+Kgbl), (p+1):(p+Kgbl)]
      Sigma.q.uGblB <- Sigma.q.betauGbl[(p+Kgbl+1):(p+2*Kgbl), (p+Kgbl+1):(p+2*Kgbl)]

      # Update lambda.q.sigsq.gbl.A and lambda.q.sigsq.gbl.B:

      lambda.q.sigsq.gbl.A <- (mu.q.recip.a.gbl.A + sum(mu.q.uGblA^2)
                               + sum(diag(Sigma.q.uGblA)))

      lambda.q.sigsq.gbl.B <- (mu.q.recip.a.gbl.B + sum(mu.q.uGblB^2)
                               + sum(diag(Sigma.q.uGblB)))

      # Update lambda.q.sigsq.eps, Lambda.q.Sigma and lambda.q.sigsq.grp:

      lambda.q.sigsq.eps <- mu.q.recip.a.eps
      Lambda.q.Sigma <- M.q.inv.A.Sigma
      lambda.q.sigsq.grp <- mu.q.recip.a.grp

      for (i in 1:m) 
      {
         mu.q.uLinGrp[[i]] <- S1$x2[[i]]
         Sigma.q.uLinGrp[[i]] <- S1$Asup22[[i]]
         E.q.betauGblLinGrp[[i]] <- S1$Asup12[[i]]

         residVec <- (yList[[i]] - CgblList[[i]]%*%mu.q.betauGbl
                               - CgrpList[[i]]%*%mu.q.uLinGrp[[i]])
         lambda.q.sigsq.eps <- lambda.q.sigsq.eps + sum(residVec^2)
         lambda.q.sigsq.eps <- lambda.q.sigsq.eps + sum(CgblList[[i]] * tcrossprod(CgblList[[i]],Sigma.q.betauGbl))
         lambda.q.sigsq.eps <- lambda.q.sigsq.eps + sum(CgrpList[[i]] * tcrossprod(CgrpList[[i]],Sigma.q.uLinGrp[[i]]))
         lambda.q.sigsq.eps <- lambda.q.sigsq.eps + 2*sum(CgrpList[[i]] * (CgblList[[i]] %*% E.q.betauGblLinGrp[[i]]))

         indsGrpLin <- 1:q
         Lambda.q.Sigma <- (Lambda.q.Sigma + tcrossprod(mu.q.uLinGrp[[i]][indsGrpLin])
                            + Sigma.q.uLinGrp[[i]][indsGrpLin,indsGrpLin])

         indsGrpSpl <- (q+1):(q+Kgrp)
         lambda.q.sigsq.grp <- (lambda.q.sigsq.grp + sum(mu.q.uLinGrp[[i]][indsGrpSpl]^2)
                                  + sum(diag(Sigma.q.uLinGrp[[i]][indsGrpSpl,indsGrpSpl])))
      }

      # Do each of the mean reciprocal and inverse updates for variance and covariance parameters:

      mu.q.recip.sigsq.eps <- xi.q.sigsq.eps/lambda.q.sigsq.eps
      mu.q.recip.sigsq.gbl.A <- xi.q.sigsq.gbl.A/lambda.q.sigsq.gbl.A
      mu.q.recip.sigsq.gbl.B <- xi.q.sigsq.gbl.B/lambda.q.sigsq.gbl.B
      M.q.inv.Sigma <- (xi.q.Sigma - q + 1)*solve(Lambda.q.Sigma)
      mu.q.recip.sigsq.grp <- xi.q.sigsq.grp/lambda.q.sigsq.grp

      # Do each of the mean reciprocal and inverse updates for auxiliary parameters:

      lambda.q.a.sigsq.eps <- mu.q.recip.sigsq.eps + (1 / (nu.eps * s.eps^2))
      mu.q.recip.a.eps <- xi.q.a.sigsq.eps / lambda.q.a.sigsq.eps

      lambda.q.a.sigsq.gbl.A <- mu.q.recip.sigsq.gbl.A + (1 / (nu.gbl.A * s.gbl.A^2))
      mu.q.recip.a.gbl.A <- xi.q.a.sigsq.gbl.A / lambda.q.a.sigsq.gbl.A

      lambda.q.a.sigsq.gbl.B <- mu.q.recip.sigsq.gbl.B + (1 / (nu.gbl.B * s.gbl.B^2))
      mu.q.recip.a.gbl.B <- xi.q.a.sigsq.gbl.B / lambda.q.a.sigsq.gbl.B

      diag(Lambda.q.A.Sigma) <- diag(M.q.inv.Sigma) +  (1/(nu.Sigma * s.Sigma^2))
      M.q.inv.A.Sigma <- xi.q.A.Sigma * solve(Lambda.q.A.Sigma)

      lambda.q.a.sigsq.grp <- mu.q.recip.sigsq.grp + (1 / (nu.grp * s.grp^2))
      mu.q.recip.a.grp <- xi.q.a.sigsq.grp / lambda.q.a.sigsq.grp

      if (itnum>=maxIter) converged <- TRUE

   }

   # Return output list:

   return(list(mu.q.betauGbl = mu.q.betauGbl,mu.q.uLinGrp = mu.q.uLinGrp,
               Sigma.q.betauGbl = Sigma.q.betauGbl,Sigma.q.uLinGrp = Sigma.q.uLinGrp,
               E.q.betauGblLinGrp = E.q.betauGblLinGrp,
               xi.q.sigsq.eps = xi.q.sigsq.eps,lambda.q.sigsq.eps = lambda.q.sigsq.eps,
               xi.q.sigsq.gbl.A = xi.q.sigsq.gbl.A,lambda.q.sigsq.gbl.A = lambda.q.sigsq.gbl.A,
               xi.q.sigsq.gbl.B = xi.q.sigsq.gbl.B,lambda.q.sigsq.gbl.B = lambda.q.sigsq.gbl.B,
               xi.q.Sigma = xi.q.Sigma,Lambda.q.Sigma = Lambda.q.Sigma,
               xi.q.sigsq.grp = xi.q.sigsq.grp,lambda.q.sigsq.grp = lambda.q.sigsq.grp,
               xi.q.A.Sigma = xi.q.A.Sigma, Lambda.q.A.Sigma = Lambda.q.A.Sigma))
}

############ End of twoLevGrpSpecCurvContrMFVB ############
