########## R script: ultrasoundExample ##########

# For performing the example in the Menictas, Nolan, Simpson & Wand 
# article in which three-level group specific curves are fitted to data 
# from an experiment involving quantitative ultrasound technology
# from the Bioacoustics Research Laboratory, Department of Electrical 
# and Computer Engineering, University of Illinois at Urbana-Champaign, 
# Illinois, U.S.A. using a streamlined Bayesian variational inference 
# approach.

# Last changed: 26 FEB 2020

# Set tasks:

doSetupForFitting <- TRUE
doFitting <- TRUE
plotFits <- TRUE
plotgFuncts <- TRUE
plothFuncts <- TRUE
plotgAndhFuncts <- TRUE

# Set the tolerance value for convergence of the mean field variational Bayes
# algorithm and the maximum numbeer of mean field variational Bayes iterations:

toleranceVal <- 1e-8
maxIterVal <- 100

# Set progress reporting flag values:

plotLogMLval <- TRUE
quietVal <- FALSE

# Source required functions:

source("ZOSull.r")
source("matrixSqrt.r")
source("SolveThreeLevelSparseLeastSquares.r")
source("threeLevGrpSpecCurvMFVB.r")
source("threeLevGrpSpecCurvLogML.r")

# Load required packages:

library(lattice)

# Set graphical parameters:

stripCexVal <- 0.6
dataCol <- "blue"
fitCol <- "darkmagenta"
ptsColour <- "dodgerblue"
shadeCol <- "thistle3"

colourVec <- c("red","deeppink","blue","pink","gold","orange","purple","navy",
               "darkmagenta","green3","brown","green4","cyan","magenta",
               "olivedrab","grey50","darkorange","sienna","limegreen",
               "chartreuse","rosybrown","green","rosybrown4","hotpink","tan",
               "cadetblue","darkviolet","mistyrose4","deepskyblue","darkgreen",
               "orange4","purple4","lavender","blueviolet","forestgreen",
               "navajowhite3","mistyrose3","darkorchid","honeydew3","wheat",
               "chocolate3","lightsteelblue4","gold3","springgreen3","lightblue4",
               "violetred4","thistle4","lightpink3","darkorange3","hotpink2")

ylabVal <- expression(paste("10 ",log[10],
               "(backscatter coefficient (centimeters",phantom()^{-1},
               "steradian",phantom()^{-1},"))"))

# Load in data and set up additional variables:

data <- read.table("ultrasound.txt",header=TRUE)
xOrig <- data$freq
yOrig <- data$BSC

# Transform x and y data as priors assume this scale:

meanx <- mean(xOrig) ; sdx <- sd(xOrig)
x <- (xOrig - meanx)/sdx

meany <- mean(yOrig) ; sdy <- sd(yOrig)
y <- (yOrig - meany)/sdy

idnumL1 <- data$idnumL1
idnumL2 <- data$idnumL2
m <- length(unique(idnumL1))
nVal <- length(unique(idnumL2))
nVec <- rep(nVal,m) # Note that the design is balanced in that
                    # the n_i values are constant.

idnumL1fac <- idnumL1
for (i in 1:m)
   idnumL1fac[idnumL1==i] <- paste("tumor",as.character(i))
idnumL1fac <- as.factor(idnumL1fac)
idnumL1fac <- ordered(idnumL1fac,paste("tumor",1:m))
idnumL2fac <- idnumL2

for (j in 1:nVal)
   idnumL2fac[idnumL2==j] <- paste("slice",as.character(j))
idnumL2fac <- as.factor(idnumL2fac)
idnumL2fac <- ordered(idnumL2fac,paste("slice",1:nVal))

if (doSetupForFitting) 
{
   uqRodent <- unique(idnumL1)
   uqSlice <- unique(idnumL2)

   numObs <- length(y)
   numRodent <- length(uqRodent)
   numSlice <- length(uqSlice)

   # Extract dimension variables

   oMat <- matrix(0,nrow=numRodent,ncol=numSlice)
   for (i in 1:m)
      for (j in 1:nVec[i])
         oMat[i,j] <- length(y[idnumL1==i & idnumL2==j])

   sumnVec <- sum(nVec)
   cumsumnVec <- c(0,cumsum(nVec))

   ind <- rep(NA,numObs)
   uqRodent <- unique(idnumL1)
   count <- 0

   for (i in 1:m) 
   {
      currRodent <- uqRodent[i]
      currInds <- (1:numObs)[idnumL1==currRodent]
      uqSlice <- unique(idnumL2[currInds])

      for (j in 1:nVec[i]) 
      {
         count <- count + 1
         ind[(idnumL1==i)&(idnumL2==j)] <- count
      }
   }

   # Set up X and Z matrices:

   X <- cbind(rep(1,numObs),x)
   ncX <- ncol(X)

   range.x <- c(1.02*min(x) - 0.02*max(x),1.02*max(x) - 0.02*min(x))

   numIntKnotsf <- 15
   numIntKnotsg <- 10
   numIntKnotsh <-  7

   intKnotsf <- quantile(unique(x),seq(0,1,length=numIntKnotsf+2))[-c(1,numIntKnotsf+2)]
   intKnotsg <- quantile(unique(x),seq(0,1,length=numIntKnotsg+2))[-c(1,numIntKnotsg+2)]
   intKnotsh <- quantile(unique(x),seq(0,1,length=numIntKnotsh+2))[-c(1,numIntKnotsh+2)]

   Zgbl <- ZOSull(x,range.x,intKnotsf)
   Zgrp1 <- ZOSull(x,range.x,intKnotsg)
   Zgrp2 <- ZOSull(x,range.x,intKnotsh)

   # Construct grid-wise design matrices for plotting:

   ng <- 101
   xg <- seq(range.x[1],range.x[2],length = ng)
   xgOrig <- meanx + sdx*xg
   Xg <- cbind(rep(1,ng),xg)
   Zgblg <- ZOSull(xg,range.x = range.x,intKnots = intKnotsf)
   Zgrp1g <- ZOSull(xg,range.x = range.x,intKnots = intKnotsg)
   Zgrp2g <- ZOSull(xg,range.x = range.x,intKnots = intKnotsh)

   # Do setup for plotting:

   xOrigList <- vector("list",m)
   yOrigList <- vector("list",m)
   for (i in 1:m)
   {
      xOrigList[[i]] <- vector("list",nVec[i])
      yOrigList[[i]] <- vector("list",nVec[i])
      for (j in 1:nVec[i])
      {
         indsi <- (1:numObs)[idnumL1==i]
         indsj <- (1:numObs)[idnumL2==j]

         indsCurr <- intersect(indsi,indsj)
         xOrigList[[i]][[j]] <- xOrig[indsCurr]
         yOrigList[[i]][[j]] <- yOrig[indsCurr]
      }
   }

   ncZgbl <- ncol(Zgbl)
   ncZgrp1 <- ncol(Zgrp1)
   ncZgrp2 <- ncol(Zgrp2)

   # Set hyperparameters for rstan:

   mu.betaHYP <- rep(0,ncX)
   Sigma.betaHYP <- 1e10*diag(ncX)
   nuHYP <- 2 ; AepsHYP <- 1e5
   ASigmaL1HYP <- 1e5
   ASigmaL2HYP <- 1e5
   AgblHYP <- 1e5
   Agrp1HYP <- 1e5
   Agrp2HYP <- 1e5

   # Set hyperparameters for MFVB:

   mu.beta <- rep(0, ncX)
   Sigma.beta <- 1e10 * diag(ncX)
   s.eps <- 1e5
   nu.eps <- 1
   s.gbl <- 1e5
   nu.gbl <- 1
   s.SigmaL1 <- rep(1e5, 2)
   nu.SigmaL1 <- 2
   s.grp1 <- 1e5
   nu.grp1 <- 1
   s.SigmaL2 <- rep(1e5, 2)
   nu.SigmaL2 <- 2
   s.grp2 <- 1e5
   nu.grp2 <- 1
}

if (doFitting) 
{
   # Set up data lists:

   yList <- vector("list",m)
   XList <- vector("list",m)
   ZgblList <- vector("list",m)
   Zgrp1List <- vector("list",m)
   Zgrp2List <- vector("list",m)

   for (i in 1:m) 
   {
      for (j in 1:nVec[i]) 
      {
         indsij <- (1:numObs)[idnumL1==i & idnumL2==j]
         yList[[i]][[j]] <- y[indsij]
         XList[[i]][[j]] <- X[indsij,]
         ZgblList[[i]][[j]] <- Zgbl[indsij,]
         Zgrp1List[[i]][[j]] <- Zgrp1[indsij,]
         Zgrp2List[[i]][[j]] <- Zgrp2[indsij,]
      }
   }

   # Call threeLevGrpSpecCurvMFVB():

   threeLevGrpSpecCurvMFVBobj <- threeLevGrpSpecCurvMFVB(
                     yList      = yList,
                     XList      = XList,
                     ZgblList   = ZgblList,
                     Zgrp1List  = Zgrp1List,
                     Zgrp2List  = Zgrp2List,
                     nVec       = nVec,
                     oMat       = oMat,
                     mu.beta    = mu.betaHYP,
                     Sigma.beta = Sigma.betaHYP,
                     s.eps      = s.eps,
                     nu.eps     = nu.eps,
                     s.gbl      = s.gbl,
                     nu.gbl     = nu.gbl,
                     s.SigmaL1  = s.SigmaL1,
                     nu.SigmaL1 = nu.SigmaL1,
                     s.SigmaL2  = s.SigmaL2,
                     nu.SigmaL2 = nu.SigmaL2,
                     s.grp1     = s.grp1,
                     nu.grp1    = nu.grp1,
                     s.grp2     = s.grp2,
                     nu.grp2    = nu.grp2,
                     maxIter    = maxIterVal,
                     tolerance  = toleranceVal,
                     plotLogML  = plotLogMLval,
                     quiet      = quietVal
                   )

   # Unpack the threeLevGrpSpecCurvMFVB() fit object:

   mu.q.betauGbl <- threeLevGrpSpecCurvMFVBobj$mu.q.betauGbl
   mu.q.uLinGrp1 <- threeLevGrpSpecCurvMFVBobj$mu.q.uLinGrp1
   mu.q.uLinGrp2 <- threeLevGrpSpecCurvMFVBobj$mu.q.uLinGrp2
   Sigma.q.betauGbl <- threeLevGrpSpecCurvMFVBobj$Sigma.q.betauGbl
   Sigma.q.uLinGrp1 <- threeLevGrpSpecCurvMFVBobj$Sigma.q.uLinGrp1
   Sigma.q.uLinGrp2 <- threeLevGrpSpecCurvMFVBobj$Sigma.q.uLinGrp2
   E.q.betauGblLinGrp1 <- threeLevGrpSpecCurvMFVBobj$E.q.betauGblLinGrp1
   E.q.betauGblLinGrp2 <- threeLevGrpSpecCurvMFVBobj$E.q.betauGblLinGrp2
   E.q.uLinGrp1uLinGrp2 <- threeLevGrpSpecCurvMFVBobj$E.q.uLinGrp1uLinGrp2
   mu.q.recip.sigsq.eps <- threeLevGrpSpecCurvMFVBobj$mu.q.recip.sigsq.eps
   mu.q.recip.sigsq.gbl <- threeLevGrpSpecCurvMFVBobj$mu.q.recip.sigsq.gbl
   M.q.inv.SigmaL1 <- threeLevGrpSpecCurvMFVBobj$M.q.inv.SigmaL1
   M.q.inv.SigmaL2 <- threeLevGrpSpecCurvMFVBobj$M.q.inv.SigmaL2
   mu.q.recip.sigsq.grp1 <- threeLevGrpSpecCurvMFVBobj$mu.q.recip.sigsq.grp1
   mu.q.recip.sigsq.grp2 <- threeLevGrpSpecCurvMFVBobj$mu.q.recip.sigsq.grp2

   fHatgMFVB <- as.vector(cbind(Xg,Zgblg)%*%mu.q.betauGbl)

   curvEstsMFVB <- vector("list",m)
   lowerMFVB <- vector("list",m)
   upperMFVB <- vector("list",m)
   gHatgMFVB <- vector("list",m)
   hHatgMFVB <- vector("list",m)
   sdMFVB <- vector("list",m)

   for (i in 1:m) 
   {
      curvEstsMFVB[[i]] <- vector("list",nVec[i])
      lowerMFVB[[i]] <- vector("list",nVec[i])
      upperMFVB[[i]] <- vector("list",nVec[i])
      hHatgMFVB[[i]] <- vector("list",nVec[i])
      sdMFVB[[i]] <- vector("list",nVec[i])
      gHatgMFVB[[i]] <- as.vector(cbind(Xg,Zgrp1g)%*%mu.q.uLinGrp1[[i]])
      for (j in 1:nVec[i]) 
         hHatgMFVB[[i]][[j]] <- as.vector(cbind(Xg,Zgrp2g)%*%mu.q.uLinGrp2[[i]][[j]])
   }

   # Transform all curves to the original scale of the data:

   fHatgMFVB <- meany +  sdy*fHatgMFVB
   for (i in 1:m) 
   {
      gHatgMFVB[[i]] <- sdy*gHatgMFVB[[i]]
      for (j in 1:nVec[i]) 
         hHatgMFVB[[i]][[j]] <- sdy*hHatgMFVB[[i]][[j]]
   }

   # Obtain combined curve estimates and pointwise 99% credible intervals: 

   for (i in 1:m) 
   {
      for (j in 1:nVec[i]) 
      {
         curvEstsMFVB[[i]][[j]] <- fHatgMFVB + gHatgMFVB[[i]] + hHatgMFVB[[i]][[j]]

         CgCg <- cbind(Xg,Zgblg,Xg,Zgrp1g,Xg,Zgrp2g)
         CovMatCurr <- rbind(cbind(Sigma.q.betauGbl,E.q.betauGblLinGrp1[[i]],
                                   E.q.betauGblLinGrp2[[i]][[j]]),
                             cbind(t(E.q.betauGblLinGrp1[[i]]),
                             Sigma.q.uLinGrp1[[i]],E.q.uLinGrp1uLinGrp2[[i]][[j]]),
                             cbind(t(E.q.betauGblLinGrp2[[i]][[j]]),
                             t(E.q.uLinGrp1uLinGrp2[[i]][[j]]),Sigma.q.uLinGrp2[[i]][[j]]))

         sdMFVB[[i]][[j]] <-  sdy*sqrt(diag(CgCg%*%CovMatCurr%*%t(CgCg)))

         lowerMFVB[[i]][[j]] <- curvEstsMFVB[[i]][[j]] - qnorm(0.995)*sdMFVB[[i]][[j]]
         upperMFVB[[i]][[j]] <- curvEstsMFVB[[i]][[j]] + qnorm(0.995)*sdMFVB[[i]][[j]]
      }
   }
}

if (plotFits)
{
    # Do the lattice plot:

    par(mfrow=c(1,1),mai=c(1.02,0.82,0.82,0.42))

    layoutVec <- c(m,nVec[1])

    tmp <- trellis.par.get("strip.background")
    tmp$col <- c("navajowhite", "palegreen1")
    trellis.par.set("strip.background",tmp)

    figFits <- xyplot(yOrig~xOrig|idnumL2fac*idnumL1fac,groups=idnumL2fac,
                         data=data.frame(x=xOrig,y=yOrig,idnumL1fac=idnumL1fac,
                                         idnumL2fac=idnumL2fac),
                         layout=layoutVec,main="",
                         strip=strip.custom(par.strip.text= list(cex = stripCexVal)),
                         par.settings = list(layout.heights = list(strip = stripCexVal)),
                         scales = list(cex = 0.7),
                         xlab="frequency",ylab=ylabVal,as.table=TRUE,
                         panel=function(x,y,subscripts,groups)
                         {
                            iPan <- panel.number()

                            i <- rep(1:m,each=nVal)[iPan]
                            j <- rep(1:nVal,m)[iPan]

                            panel.grid()

                            panel.polygon(c(xgOrig,rev(xgOrig)),
                                          c(lowerMFVB[[i]][[j]],rev(upperMFVB[[i]][[j]])),
                                          col="lightcoral",border=FALSE)

                            panel.xyplot(xgOrig,curvEstsMFVB[[i]][[j]],
                                         col="red",type="l",lwd=2)

                            panel.superpose(x[order(x)],y[order(x)],subscripts,groups,
                                        type="l",col=dataCol)

                         })

    print(figFits)

    readline("Hit Enter to continue.\n")
}

if (plotgFuncts)
{
   par(mfrow=c(1,1),mai=c(1.02,0.82,0.82,0.42))

   ylimVal <- range(unlist(gHatgMFVB))

   mainString <- expression(paste("estimated ",g[i]," functions"))
   plot(0,type="n",xlim=range(xgOrig),ylim=ylimVal,bty="l",xlab="frequency",ylab=ylabVal,
        main=mainString,cex.main=1.5)

   for (i in 1:m)
      lines(xgOrig,gHatgMFVB[[i]],col=colourVec[i],lwd=2)

   abline(h=0,col="black",lwd=4)

   readline("Hit Enter to continue.\n")
}

if (plothFuncts)
{
   par(mfrow=c(1,1),mai=c(1.02,0.82,0.82,0.42))

   ylimVal <- range(unlist(hHatgMFVB))
   mainString <- expression(paste("estimated ",h[ij]," functions"))
   plot(0,type="n",xlim=range(xgOrig),ylim=ylimVal,bty="l",xlab="frequency",ylab=ylabVal,
        main=mainString,cex.main=1.5)

   icolour <- 0
   for (i in 1:m)
      for (j in 1:nVal)
      {
         icolour <- icolour + 1
         lines(xgOrig,hHatgMFVB[[i]][[j]],col=colourVec[icolour],lwd=2)
      } 

   abline(h=0,col="black",lwd=4)
   readline("Hit Enter to continue.\n")
}

if (plotgAndhFuncts)
{
   gColour <- "darkblue"
   hColour <- "dodgerblue"

   par(mfrow=c(2,5),mai=c(0.3,0.3,0.5,0.05))
   for (i in 1:m)
   {
       plot(0,type="n",xlim=range(xgOrig),ylim=ylimVal,bty="l",xlab="frequency",ylab="",
            main=paste("tumor",i),cex.main=1.2)
       lines(xgOrig,gHatgMFVB[[i]],col=gColour,lwd=2)
       for (j in 1:nVal)
         lines(xgOrig,hHatgMFVB[[i]][[j]],col=hColour,lwd=1)
      
      abline(h=0,col="black",lwd=3)

      if (i==1)
         legend("topleft",legend=c(expression(g[i]),expression(h[ij])),
                col=c(gColour,hColour),lwd=c(2,1))
   }
}

############ End of ultrasoundExample ############
